#!/usr/bin/env python
#This program will retrieve all objects definitions stored on a bioMoby Server 
#and generate the corresponding Python Object

#We need a connection to bioMoby
from bioMoby.mobyClient import Client
#some predefined types
from bioMoby.mobyDataTypes import MobyObject, MobyString
#and a model for xml
from xml.dom.minidom import *
#some regular expression to clean the xml
import re
import string

import sys

#the intern attributes
attributes=['namespace', 'id']

paramDict={}
paramDict['Object']=[("namespace","\"\"",""),("id","\"\"","")]
paramDict['String']=[("content","\"\"",""),("namespace","\"\"",""),("id","\"\"","")]

def cleanName(aName):
    """ Cleaning the name of the object
    """
    if aName=="":
        return ""

    if aName=="namespace" or aName=="id" or aName=='content':
        return aName        

    r=re.compile("\W")
    
    t=r.split(aName)
    
    for i in range(0,len(t)):
        t[i]=string.upper(t[i][0])+t[i][1:]

    return "".join(t)

def makeMobyName(aName):
    """insert a Moby prefix for the name of the object
    """
    mobyName='Moby'+cleanName(aName)

    return mobyName    

def makeFromMoby(parameters):
    """Generate the deserializers
    """

    res=["    def fromMoby(self, value):\n"]
    res.append("        \"\"\" Load the attributes from an XML string\n        \"\"\"")
    res.append("        from bioMoby.mobyMarshal import MobyUnmarshaller")
    res.append("        doc=value")
    res.append("        if isinstance(value,str):")
    res.append("            doc=parseString(value).firstChild")
    res.append("        um=MobyUnmarshaller()")
    res.append("        self.content=\"\"")
    
    parNames=[]

    for i in parameters:
        parNames.append(i[0])

    for att in attributes:
        if att in parNames:
            res.append("        self."+att+"=doc.getAttribute(\"moby:"+att+"\")")
            res.append("        if self."+att+"==\"\":")
            res.append("            self."+att+"=doc.getAttribute(\""+att+"\")")
            
    res.append("\n        self.articleName=doc.getAttribute(\"moby:articleName\")")
    res.append("        if self.articleName==\"\":\n")
    res.append("            self.articleName=doc.getAttribute(\"articleName\")")
    

    for par in parameters:
        if par[1]=="[]":
            res.append("        self."+cleanName(par[0])+"=[]")

    res.append("        for elt in doc.childNodes:")
    res.append("            if elt.nodeType==elt.TEXT_NODE or elt.nodeType==elt.CDATA_SECTION_NODE:")
    res.append("                self.content=self.content+elt.nodeValue")
    res.append("            if elt.nodeName==\"CrossReference\" or elt.nodeName==\"moby:CrossReference\":")
    res.append("                self.__cross__=[]")
    res.append("                for ref in elt.childNodes:")
    res.append("                    self.__cross__.append(ref)")        

    for parameter in parameters:
        if parameter[0] not in attributes and parameter[0]!="content":
            res.append("            if elt.nodeType==elt.ELEMENT_NODE and (elt.nodeName==\""+parameter[2]+"\" or elt.nodeName==\"moby:"+parameter[2]+"\") and (elt.getAttribute(\"articleName\")==\""+cleanName(parameter[0])+"\" or elt.getAttribute(\"moby:articleName\")==\""+cleanName(parameter[0])+"\"):")
            
            if parameter[2]=='Integer' or parameter[2]=='Float' or parameter[2]=='String':
                res.append("                t="+makeMobyName(parameter[2])+"()")
                res.append("                t.fromMoby(elt)")
                if parameter[1]!="[]":
                    res.append("                self."+cleanName(parameter[0])+"=t.content")
                else:
                    res.append("                self."+cleanName(parameter[0])+".append(t.content)")
                    
            elif parameter[0]!='content':
                if parameter[1]!="[]":
                    res.append("                self."+cleanName(parameter[0])+"="+makeMobyName(parameter[2])+"()")
                    res.append("                self."+cleanName(parameter[0])+".fromMoby(elt)")
                else:
                    res.append("                auxObject="+makeMobyName(parameter[2])+"()")
                    res.append("                auxObject.fromMoby(elt)")
                    res.append("                self."+cleanName(parameter[0])+".append(auxObject)")
                
    return "\n".join(res)

def getPropertiesFromAncestor(ancestor):
    """Get the attributes from the ancestor
    """
    prop=[]
    for param in paramDict[ancestor]:
        prop.append(param)
    return prop

def makeInterface(objectName, path=""):
    """make the interface from the Moby Definition
    """
    if objectName in paramDict.keys():
        return
        
    objectDefinition=client.retrieveObjectDefinition(objectName)
    print "Creating interface for: "+objectName+"..."
    
    #We retrieve the relationships
    properties=[]
    ancestor=""
    
    #we begin to build our interface
    if "ISA" in objectDefinition["Relationship"]:
        ancestor=objectDefinition["Relationship"]["ISA"][0][1]
        ancestor=ancestor.split(":")
        ancestor=ancestor[len(ancestor)-1]

        if ancestor not in paramDict.keys():
            makeInterface(ancestor, path)
	
        for aProp in getPropertiesFromAncestor(ancestor):
            properties.append(aProp)
   
    
    if "HASA" in objectDefinition["Relationship"].keys():
        for hasaRelationship in objectDefinition["Relationship"]["HASA"]:
            articleName=hasaRelationship[0]
            if articleName=="":
                articleName=objectName

            t0=hasaRelationship[1]
            t1=t0.split(":")
            articleType=t1[len(t1)-1]
                
            properties.append((articleName,"None",articleType))

	    
    if "HAS" in objectDefinition["Relationship"].keys():
        for hasRelationship in objectDefinition["Relationship"]["HAS"]:
            articleName=hasRelationship[0]
            if articleName=="":
                articleName=objectName

            t0=hasRelationship[1]
            t1=t0.split(":")
            articleType=t1[len(t1)-1]
                
            properties.append((articleName,"[]",articleType))

    #now we can build our python interface:

    mAncestor=makeMobyName(ancestor)
    mObjectName=makeMobyName(objectName)

    
    paramDict[objectName]=properties

    pyResult=["#Autogenerated class\n"]
    
    pyResult.append("\n\nfrom xml.dom.minidom import parseString\n\n")

    already=[]

    for property in properties:
        if property[2] not in already:
            if property[2]=='Object' or property[2]=='String' or property[2]=='Integer' or property[2]=='Float':
                pyResult.append("from bioMoby.mobyDataTypes import Moby"+property[2]+"\n")
            elif property[2]!="":
                pyResult.append("from "+cleanName(property[2])+" import "+makeMobyName(property[2])+"\n")
            already.append(property[2])
    
    if ancestor!="":
        if (ancestor=='Object' or ancestor=='String' or ancestor=='Integer' or ancestor=='Float'):
            if ancestor not in already:
                pyResult.append("from bioMoby.mobyDataTypes import "+mAncestor+"\n")
        else:
            pyResult.append("from "+cleanName(ancestor)+" import "+mAncestor+"\n")
        pyResult.append("\n\nclass "+mObjectName+"("+mAncestor+"):\n")
    else:
        pyResult.append("\n")
        pyResult.append("class "+mObjectName+":\n")

    if "Description" in objectDefinition:
	description=objectDefinition["Description"]
    if "authURI" in objectDefinition:
	author=objectDefinition["authURI"]
    if "contactEmail" in objectDefinition:
        contact=objectDefinition["contactEmail"]
    
    pyResult.append("    \"\"\" Description: "+description+"\n")
    pyResult.append("        Author: "+author+"\n")
    pyResult.append("        Contact: "+contact+"\n")
    pyResult.append("    \"\"\"\n")
                
    initstr=["    def __init__(self"]

    for property in properties:
        if property[1]!="[]" and property[2]=='String':
            initstr.append(","+cleanName(property[0])+"=\"\"")
        elif property[1]!="[]" and property[2]=='Integer':
            initstr.append(","+cleanName(property[0])+"=0")
        elif property[1]!="[]" and property[2]=='Float':
            initstr.append(","+cleanName(property[0])+"=0.0")
        else:
            initstr.append(","+cleanName(property[0])+"="+property[1])
    initstr.append(",articleName=\"\"):\n")

    pyResult.append("".join(initstr))

    articles=["        self._articles={"]
    t1=[]
    for property in properties:
        if property[0] != 'namespace' and property[0] != 'id' and property[0] != 'content':
            t1.append("\""+cleanName(property[0])+"\":\""+property[0]+"\"")
    articles.append(",".join(t1))
    
    articles.append("}\n")
    
    pyResult.append("".join(articles))

    for property in properties:
        pyResult.append("        self."+cleanName(property[0])+"="+cleanName(property[0])+"\n")
        
    pyResult.append("        self.__prefix__=\"moby\"\n")
    pyResult.append("        self.__tag__=\""+objectName+"\"\n")
    pyResult.append("        self.articleName=articleName\n")
    pyResult.append("\n\n")

    pyResult.append(makeFromMoby(paramDict[objectName]))
        
    fp=file(path+"/"+cleanName(objectName)+".py","w")
    fp.writelines(pyResult)
    fp.close()    

""" The main program
     First setup a connection to the bioMoby server
     first argument is the name of the directory where to put your objects
     other arguments are names of objects you wanted to be included in your ontology
"""

if __name__ == '__main__':
    outputpath="output"

    if len(sys.argv)>=2:
        outputpath=sys.argv[1]
    
    #Connection to mobyServer:
    client=Client()

    #first the program retrieves the object names from bioMoby
    print "Retrieve all object names on bioMoby..."
    objectNames=client.retrieveObjectNames().keys()

    #delete from the list Object and String:
    objectNames.remove("Object")
    objectNames.remove("String")
    objectNames.remove("Float")
    objectNames.remove("Integer")

    if len(sys.argv)>2:
        objectNames=sys.argv[2:]
    #for each objectName, it retrieves the object's definition
    for objectName in objectNames:
        makeInterface(objectName,path=outputpath)
    
    import os

    ld=os.listdir(outputpath)

    toRead=[]

    for f in ld:
        if f[-4:]!='.pyc':
            toRead.append(f)
    try:        
        toRead.remove("__init__.py")
        toRead.remove("CVS")
    except:
        pass

    initfile=["#All classes are declared in this generated file\n"]
    initfile.append("\n")

    for f in toRead:
        modulename=f.split('.')[0]
        initfile.append("from "+cleanName(modulename)+" import "+makeMobyName(modulename)+"\n")
    
    fp=file(outputpath+"/__init__.py",'w')
    fp.writelines(initfile)
    fp.close()
